(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[662],{79655:function(n,t,u){(window.__NEXT_P=window.__NEXT_P||[]).push(["/settings",function(){return u(16470)}])},16470:function(n,t,u){"use strict";u.r(t),u.d(t,{default:function(){return f}});var e=u(67294),c=u(11163);function f(){let n=(0,c.useRouter)();return(0,e.useEffect)(()=>{n.replace("/settings/config")},[n]),null}}},function(n){n.O(0,[888,774,179],function(){return n(n.s=79655)}),_N_E=n.O()}]);